from __future__ import absolute_import
from socket import error as socket_error
from decimal import Decimal

from celery import shared_task
from celery.utils.log import get_task_logger
from bitcoinrpc.authproxy import AuthServiceProxy

from django.db import transaction

from cc.models import (Wallet, Currency, Transaction, Address,
                       WithdrawTransaction, Operation)
from cc import settings

logger = get_task_logger(__name__)


@shared_task(throws=(socket_error,))
@transaction.atomic
def query_transactions(ticker=None):
    if not ticker:
        for c in Currency.objects.all():
            query_transactions.delay(c.ticker)
        return

    currency = Currency.objects.select_for_update().get(ticker=ticker)
    coin = AuthServiceProxy(currency.api_url)
    current_block = coin.getblockcount()
    processed_transactions = []

    block_hash = coin.getblockhash(currency.last_block)
    transactions = coin.listsinceblock(block_hash)['transactions']

    for tx in transactions:
        if tx['txid'] in processed_transactions:
            continue

        if tx['category'] not in ('receive', 'generated'):
            continue

        process_deposite_transaction(tx, ticker)
        processed_transactions.append(tx['txid'])

    currency.last_block = current_block
    currency.save()

    for tx in Transaction.objects.filter(processed=False, currency=currency):
        query_transaction(ticker, tx.txid)


@transaction.atomic
def process_deposite_transaction(txdict, ticker):
    if txdict['category'] not in ('receive', 'generated'):
        return

    try:
        address = Address.objects.select_for_update().get(address=txdict['address'])
    except Address.DoesNotExist:
        return

    currency = Currency.objects.get(ticker=ticker)

    try:
        wallet = Wallet.objects.select_for_update().get(addresses=address)
    except Wallet.DoesNotExist:
        wallet, created = Wallet.objects.select_for_update().get_or_create(
            currency=currency,
            label='_unknown_wallet'
        )
        address.wallet = wallet
        address.save()

    tx, created = Transaction.objects.select_for_update().get_or_create(txid=txdict['txid'], currency=currency)

    if tx.processed:
        return

    if created:
        if txdict['confirmations'] >= settings.CC_CONFIRMATIONS:
            Operation.objects.create(
                wallet=wallet,
                balance=txdict['amount'],
                description='Deposite',
                reason=tx
            )
            wallet.balance += txdict['amount']
            wallet.save()
            tx.processed = True
        else:
            Operation.objects.create(
                wallet=wallet,
                unconfirmed=txdict['amount'],
                description='Unconfirmed',
                reason=tx
            )
            wallet.unconfirmed += txdict['amount']
            wallet.save()

    else:
        if txdict['confirmations'] >= settings.CC_CONFIRMATIONS:
            Operation.objects.create(
                wallet=wallet,
                unconfirmed=-txdict['amount'],
                balance=txdict['amount'],
                description='Confirmed',
                reason=tx
            )
            wallet.unconfirmed -= txdict['amount']
            wallet.balance += txdict['amount']
            wallet.save()
            tx.processed = True

    tx.save()


@shared_task(throws=(socket_error,))
@transaction.atomic
def query_transaction(ticker, txid):
    currency = Currency.objects.select_for_update().get(ticker=ticker)
    coin = AuthServiceProxy(currency.api_url)
    txdict = normalise_txifno(coin.gettransaction(txid))
    process_deposite_transaction(txdict, ticker)


def normalise_txifno(data):
    txdict = data['details'][0]
    txdict['confirmations'] = data['confirmations']
    txdict['txid'] = data['txid']
    txdict['timereceived'] = data['timereceived']
    txdict['time'] = data['time']
    return txdict


@shared_task(throws=(socket_error,))
@transaction.atomic
def refill_addresses_queue():
    for currency in Currency.objects.all():
        coin = AuthServiceProxy(currency.api_url)
        count = Address.objects.filter(currency=currency, active=True, wallet=None).count()

        if count < settings.CC_ADDRESS_QUEUE:
            for i in xrange(count, settings.CC_ADDRESS_QUEUE):
                Address.objects.create(address=coin.getnewaddress(), currency=currency)


@shared_task(throws=(socket_error,))
@transaction.atomic
def process_withdraw_transacions(ticker=None):
    if not ticker:
        for c in Currency.objects.all():
            process_withdraw_transacions.delay(c.ticker)
        return

    currency = Currency.objects.select_for_update().get(ticker=ticker)
    coin = AuthServiceProxy(currency.api_url)

    wtxs = WithdrawTransaction.objects.select_for_update().select_related('wallet').filter(currency=currency, txid=None)

    transaction_hash = {}
    for tx in wtxs:
        if tx.address in transaction_hash:
            transaction_hash[tx.address] += tx.amount
        else:
            transaction_hash[tx.address] = tx.amount
    if not transaction_hash:
        return

    txid = coin.sendmany("", transaction_hash)

    if not txid:
        return

    fee = coin.gettransaction(txid).get('fee', 0) * -1
    if not fee:
        fee_per_tx = 0
    else:
        fee_per_tx = (fee / len(wtxs)).quantize(Decimal("0.00000001"))

    for tx in wtxs:
        Operation.objects.create(
            wallet=tx.wallet,
            holded=-tx.amount,
            balance=-fee_per_tx,
            description='Withdraw',
            reason=tx
        )

        wallet = Wallet.objects.get(id=tx.wallet.id)
        wallet.balance -= fee_per_tx
        wallet.holded -= tx.amount
        wallet.save()

    wtxs.update(txid=txid, fee=fee_per_tx)



_json = __import__('json')
loads = _json.loads
dumps = _json.dumps
if hasattr(_json, 'JSONEncodeException'):
	JSONEncodeException = _json.JSONEncodeException
	JSONDecodeException = _json.JSONDecodeException
else:
	JSONEncodeException = TypeError
	JSONDecodeException = ValueError



"""Validate bitcoin/altcoin addresses

Copied from:
http://rosettacode.org/wiki/Bitcoin/address_validation#Python
"""

import string
from hashlib import sha256

digits58 = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'

def _bytes_to_long(bytestring, byteorder):
    """Convert a bytestring to a long

    For use in python version prior to 3.2
    """
    result = []
    if byteorder == 'little':
        result = (v << i * 8 for (i, v) in enumerate(bytestring))
    else:
        result = (v << i * 8 for (i, v) in enumerate(reversed(bytestring)))
    return sum(result)

def _long_to_bytes(n, length, byteorder):
    """Convert a long to a bytestring

    For use in python version prior to 3.2
    Source:
    http://bugs.python.org/issue16580#msg177208
    """
    if byteorder == 'little':
        indexes = range(length)
    else:
        indexes = reversed(range(length))
    return bytearray((n >> i * 8) & 0xff for i in indexes)

def decode_base58(bitcoin_address, length):
    """Decode a base58 encoded address

    This form of base58 decoding is bitcoind specific. Be careful outside of
    bitcoind context.
    """
    n = 0
    for char in bitcoin_address:
        try:
            n = n * 58 + digits58.index(char)
        except:
            msg = u"Character not part of Bitcoin's base58: '%s'"
            raise ValueError(msg % (char,))
    try:
        return n.to_bytes(length, 'big')
    except AttributeError:
        # Python version < 3.2
        return _long_to_bytes(n, length, 'big')

def encode_base58(bytestring):
    """Encode a bytestring to a base58 encoded string
    """
    # Count zero's
    zeros = 0
    for i in range(len(bytestring)):
        if bytestring[i] == 0:
            zeros += 1
        else:
            break
    try:
        n = int.from_bytes(bytestring, 'big')
    except AttributeError:
        # Python version < 3.2
        n = _bytes_to_long(bytestring, 'big')
    result = ''
    (n, rest) = divmod(n, 58)
    while n or rest:
        result += digits58[rest]
        (n, rest) = divmod(n, 58)
    return zeros * '1' + result[::-1]  # reverse string

def validate(bitcoin_address, magicbyte=0):
    """Check the integrity of a bitcoin address

    Returns False if the address is invalid.
    >>> validate('1AGNa15ZQXAZUgFiqJ2i7Z2DPU2J6hW62i')
    True
    >>> validate('')
    False
    """
    if isinstance(magicbyte, int):
        magicbyte = (magicbyte,)
    clen = len(bitcoin_address)
    if clen < 27 or clen > 35: # XXX or 34?
        return False
    allowed_first = tuple(string.digits)
    try:
        bcbytes = decode_base58(bitcoin_address, 25)
    except ValueError:
        return False
    # Check magic byte (for other altcoins, fix by Frederico Reiven)
    for mb in magicbyte:
        if bcbytes.startswith(chr(int(mb))):
            break
    else:
        return False
    # Compare checksum
    checksum = sha256(sha256(bcbytes[:-4]).digest()).digest()[:4]
    if bcbytes[-4:] != checksum:
        return False
    # Encoded bytestring should be equal to the original address,
    # for example '14oLvT2' has a valid checksum, but is not a valid btc
    # address
    return bitcoin_address == encode_base58(bcbytes)




"""
  Copyright 2011 Jeff Garzik

  AuthServiceProxy has the following improvements over python-jsonrpc's
  ServiceProxy class:

  - HTTP connections persist for the life of the AuthServiceProxy object
    (if server supports HTTP/1.1)
  - sends protocol 'version', per JSON-RPC 1.1
  - sends proper, incrementing 'id'
  - sends Basic HTTP authentication headers
  - parses all JSON numbers that look like floats as Decimal
  - uses standard Python json lib

  Previous copyright, from python-jsonrpc/jsonrpc/proxy.py:

  Copyright (c) 2007 Jan-Klaas Kollhof

  This file is part of jsonrpc.

  jsonrpc is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This software is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this software; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

try:
    import http.client as httplib
except ImportError:
    import httplib
import base64
import json
import decimal
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse

USER_AGENT = "AuthServiceProxy/0.1"

HTTP_TIMEOUT = 30


class JSONRPCException(Exception):
    def __init__(self, rpc_error):
        Exception.__init__(self)
        self.error = rpc_error


def EncodeDecimal(o):
    if isinstance(o, decimal.Decimal):
        return round(o, 8)
    raise TypeError(repr(o) + " is not JSON serializable")

class AuthServiceProxy(object):
    __id_count = 0

    def __init__(self, service_url, service_name=None, timeout=HTTP_TIMEOUT, connection=None):
        self.__service_url = service_url
        self.__service_name = service_name
        self.__url = urlparse.urlparse(service_url)
        if self.__url.port is None:
            port = 80
        else:
            port = self.__url.port
        (user, passwd) = (self.__url.username, self.__url.password)
        try:
            user = user.encode('utf8')
        except AttributeError:
            pass
        try:
            passwd = passwd.encode('utf8')
        except AttributeError:
            pass
        authpair = user + b':' + passwd
        self.__auth_header = b'Basic ' + base64.b64encode(authpair)
        
        if connection: 
            # Callables re-use the connection of the original proxy 
            self.__conn = connection
        elif self.__url.scheme == 'https':
            self.__conn = httplib.HTTPSConnection(self.__url.hostname, port,
                                                  None, None, False,
                                                  timeout)
        else:
            self.__conn = httplib.HTTPConnection(self.__url.hostname, port,
                                                 False, timeout)

    def __getattr__(self, name):
        if name.startswith('__') and name.endswith('__'):
            # Python internal stuff
            raise AttributeError
        if self.__service_name is not None:
            name = "%s.%s" % (self.__service_name, name)
        return AuthServiceProxy(self.__service_url, name, connection=self.__conn)

    def __call__(self, *args):
        AuthServiceProxy.__id_count += 1

        postdata = json.dumps({'version': '1.1',
                               'method': self.__service_name,
                               'params': args,
                               'id': AuthServiceProxy.__id_count}, default=EncodeDecimal)
        self.__conn.request('POST', self.__url.path, postdata,
                            {'Host': self.__url.hostname,
                             'User-Agent': USER_AGENT,
                             'Authorization': self.__auth_header,
                             'Content-type': 'application/json'})

        response = self._get_response()
        if response['error'] is not None:
            raise JSONRPCException(response['error'])
        elif 'result' not in response:
            raise JSONRPCException({
                'code': -343, 'message': 'missing JSON-RPC result'})
        else:
            return response['result']

    def _batch(self, rpc_call_list):
        postdata = json.dumps(list(rpc_call_list), default=EncodeDecimal)
        self.__conn.request('POST', self.__url.path, postdata,
                            {'Host': self.__url.hostname,
                             'User-Agent': USER_AGENT,
                             'Authorization': self.__auth_header,
                             'Content-type': 'application/json'})

        return self._get_response()

    def _get_response(self):
        http_response = self.__conn.getresponse()
        if http_response is None:
            raise JSONRPCException({
                'code': -342, 'message': 'missing HTTP response from server'})

        return json.loads(http_response.read().decode('utf8'),
                          parse_float=decimal.Decimal)



# -*- coding: utf-8 -*-

from bitcoinaddress import validate
from django.db import models
from django.db.models import Sum

from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.utils.timezone import now

from cc import settings


class Wallet(models.Model):
    currency = models.ForeignKey('Currency')
    balance = models.DecimalField(_('Balance'), max_digits=18, decimal_places=8, default=0)
    holded = models.DecimalField(_('Holded'), max_digits=18, decimal_places=8, default=0)
    unconfirmed = models.DecimalField(_('Unconfirmed'), max_digits=18, decimal_places=8, default=0)
    label = models.CharField(_('Label'), max_length=100, blank=True, null=True)

    def __unicode__(self):
        return u'{0} {1} "{2}"'.format(self.balance, self.currency.ticker, self.label or '')

    def get_address(self):
        active = Address.objects.filter(wallet=self, active=True, currency=self.currency)[:1]
        if active:
            return active[0]

        unused = Address.objects.filter(wallet=None, active=True, currency=self.currency)[:1]
        if unused:
            free = unused[0]
            free.wallet = self
            free.save()
            return free

        old = Address.objects.filter(wallet=self, active=False, currency=self.currency)[:1]
        if old:
            return old[0]

    def withdraw(self, address, amount, description=""):
        if not validate(address, magicbyte=self.currency.magicbyte.split(',')):
            raise ValueError('Invalid address')

        if amount < 0:
            raise ValueError('Invalid amount')

        if self.balance - amount < -settings.CC_ALLOW_NEGATIVE_BALANCE:
            raise ValueError('No money')

        tx = WithdrawTransaction.objects.create(
            currency=self.currency,
            amount=amount,
            address=address,
            wallet=self
        )
        Operation.objects.create(
            wallet=self,
            balance=-amount,
            holded=amount,
            description=description,
            reason=tx
        )
        self.balance -= amount
        self.holded += amount
        self.save()

    def recalc_balace(self):
        return Operation.objects.filter(wallet=self).aggregate(balance=Sum('balance'),
                                                            holded=Sum('holded'),
                                                            unconfirmed=Sum('unconfirmed'))

    def get_operations(self):
        return Operation.objects.filter(wallet=self).order_by('-created')


class Operation(models.Model):
    wallet = models.ForeignKey(Wallet)
    created = models.DateTimeField(_('Created'), default=now)
    balance = models.DecimalField(_('Balance'), max_digits=18, decimal_places=8, default=0)
    holded = models.DecimalField(_('Holded'), max_digits=18, decimal_places=8, default=0)
    unconfirmed = models.DecimalField(_('Unconfirmed'), max_digits=18, decimal_places=8, default=0)

    description = models.CharField(_('Description'), max_length=100, blank=True, null=True)

    reason_content_type = models.ForeignKey(ContentType, null=True, blank=True)
    reason_object_id = models.PositiveIntegerField(null=True, blank=True)
    reason = GenericForeignKey('reason_content_type', 'reason_object_id')


class Address(models.Model):
    address = models.CharField(_('Address'), max_length=50, primary_key=True)
    currency = models.ForeignKey('Currency')
    created = models.DateTimeField(_('Created'), default=now)
    active = models.BooleanField(_('Active'), default=True)
    label = models.CharField(_('Label'), max_length=50, blank=True, null=True, default=None)
    wallet = models.ForeignKey(Wallet, blank=True, null=True, related_name="addresses")

    def __unicode__(self):
        return u'{0}, {1}'.format(self.address, self.currency.ticker)


class Currency(models.Model):
    ticker = models.CharField(_('Ticker'), max_length=4, default='BTC', primary_key=True)
    label = models.CharField(_('Label'), max_length=20, default='Bitcoin', unique=True)
    magicbyte = models.CommaSeparatedIntegerField(_('Magicbytes'), max_length=10, default='0,5')
    last_block = models.PositiveIntegerField(_('Last block'), blank=True, null=True, default=0)
    api_url = models.CharField(_('API hostname'), default='http://localhost:8332', max_length=100, blank=True, null=True)

    class Meta:
        verbose_name_plural = _('currencies')

    def __unicode__(self):
        return self.label


class Transaction(models.Model):
    txid = models.CharField(_('Txid'), max_length=100, unique=True)
    currency = models.ForeignKey('Currency')
    processed = models.BooleanField(_('Processed'), default=False)


class WithdrawTransaction(models.Model):
    currency = models.ForeignKey('Currency')
    amount = models.DecimalField(_('Amount'), max_digits=18, decimal_places=8)
    address = models.CharField(_('Address'), max_length=50)
    wallet = models.ForeignKey(Wallet)
    created = models.DateTimeField(_('Created'), default=now)
    txid = models.CharField(_('Txid'), max_length=100, blank=True, null=True)
    fee = models.DecimalField(_('Amount'), max_digits=18, decimal_places=8, null=True, blank=True)



from setuptools import setup, find_packages
from os.path import join, dirname


setup(
    name='django-cc',
    version='0.0.1',
    license='MIT License',
    description='Django wallet for Bitcoin and other cryptocurrencies',
    author='Ivan Vershigora',
    author_email='ivan.vershigora@gmail.com',
    url='https://github.com/limpbrains/django-cc',
    download_url = 'https://github.com/limpbrains/django-cc/tarball/0.0.1',
    keywords='bitcoin litecoin django wallet',
    packages=find_packages(),
    long_description=open(join(dirname(__file__), 'README.md')).read(),
    install_requires=[
        # TODO add python-bitcoinaddress and python-bitcoinrpc when new version will be released
        'Django>=1.7',
        'celery'
    ]
)
